import React, {useCallback, useContext} from "react";
import {Smartable, useTable} from "./Smartable";
import {ColumnFilter} from "./Columns/ColumnFilter";
import {SortState} from "./Sort";

/**
 * Basic column key type.
 */
export type ColumnKey = string|symbol;

/**
 * Smartable columns definition.
 */
export type Columns<K extends ColumnKey> = Record<K, Column>;

/**
 * Define columns for a Smartable.
 * @param columns
 */
export function createColumns<K extends ColumnKey>(...columns: [K, Column][]): Columns<K>
{
	return Object.fromEntries(
		columns,
	) as Columns<K>;
}

/**
 * Smartable column definition.
 */
export interface Column<T = any>
{
	/**
	 * Column title element.
	 */
	title: React.ReactNode;

	/**
	 * Column cell default element.
	 */
	cellElement?: React.ReactElement;

	/**
	 * Cells data comparison in the column.
	 * @param a First data to compare.
	 * @param b Second data to compare.
	 */
	compare?: (a: T, b: T) => number;

	/**
	 * Column filter definition.
	 */
	filter?: ColumnFilter;
}

/**
 * Create a column for a Smartable.
 * @param key The column key, which should be unique per Smartable instance.
 * @param column The column definition.
 */
export function createColumn<K extends ColumnKey>(key: K, column: Column): [K, Column]
{
	return [key, column];
}

/**
 * Table column context data.
 */
export interface ColumnContextData<CK extends ColumnKey>
{
	/**
	 * Column key.
	 */
	key: ColumnKey;

	/**
	 * Column definition.
	 */
	column: Column;

	/**
	 * Column sort state.
	 */
	sortState?: SortState;

	/**
	 * Column filter state.
	 */
	filterState: any;

	/**
	 * Set current column filter state.
	 * @param filterState New filter state.
	 */
	setFilterState: <T = any>(filterState: T) => void;
}

export const ColumnContext = React.createContext<ColumnContextData<ColumnKey>>(undefined);

/**
 * Hook to get current column data.
 */
export function useColumn<CK extends ColumnKey>(smartable?: Smartable<CK>): ColumnContextData<CK>
{
	return useContext(ColumnContext);
}

/**
 * Auto column context provider from table context.
 */
export function AutoColumnContextProvider({columnKey, children}: React.PropsWithChildren<{columnKey: ColumnKey;}>)
{
	// Get table data.
	const table = useTable();

	// Initialize filterState dispatcher for the current column.
	const setFilterState = useCallback((filterState: any) => (
		// Set the filter state for the current column key.
		table.setColumnFilterState(columnKey, filterState)
	), [columnKey, table.setColumnFilterState]);

	return (
		<ColumnContext.Provider value={{
			key: columnKey,
			// Get current column data from table data.
			column: table.columns[columnKey],
			// Get current column sort state from table data.
			sortState: table.columnsSortState?.[columnKey],
			// Get current column filter state from table data.
			filterState: table.columnsFilterStates?.[columnKey],
			// Current column filter state dispatcher.
			setFilterState: setFilterState,
		}}>
			{children}
		</ColumnContext.Provider>
	)
}
